from django.db import models
from validators import FlashcardMask, OverlapIntervalException


class MaskField(models.Field):
    def __init__(self, blank_sep=',', range_sep='-', *args, **kwargs):
        self.blank_sep = blank_sep
        self.range_sep = range_sep
        super(MaskField, self).__init__(*args, **kwargs)

    @staticmethod
    def _using_array(connection):
        return connection.settings_dict['ENGINE'] == 'django.db.backends.postgresql_psycopg2'

    def deconstruct(self):
        name, path, args, kwargs = super(MaskField, self).deconstruct()
        kwargs['blank_sep'] = self.blank_sep
        kwargs['range_sep'] = self.range_sep
        return name, path, args, kwargs

    def db_type(self, connection):
        return 'integer[2][]' if self._using_array(connection) else 'varchar'

    def from_db_value(self, value, expression, connection, context):
        if value is None:
            return value
        if self._using_array(connection):
            return MaskField._psql_parse_mask(value)
        return MaskField._varchar_parse_mask(value)

    def get_db_prep_value(self, value, connection, prepared=False):
        if not prepared:
            value = self.get_prep_value(value)
        if value is None:
            return value
        if self._using_array(connection):
            return value
        return ','.join(['-'.join(map(str, i)) for i in value])

    def to_python(self, value):
        if value is None:
            return value
        return FlashcardMask(value)

    def get_prep_value(self, value):
        if value is None:
            return value
        return sorted(map(list, FlashcardMask(value)))

    def get_prep_lookup(self, lookup_type, value):
        raise TypeError("Lookup not supported for MaskField")

    @staticmethod
    def _parse_mask(intervals):
        p_beg, p_end = -1, -1
        mask_list = []
        for interval in intervals:
            beg, end = map(int, interval)
            if not (0 <= beg <= 255) or not (0 <= end <= 255) or not (beg <= end) or not (beg > p_end):
                raise ValueError("Invalid range offsets in the mask")
            mask_list.append([beg, end])
            p_beg, p_end = beg, end
        return mask_list

    @classmethod
    def _varchar_parse_mask(cls, value):
        if not value:
            return FlashcardMask([])

        mask = [tuple(map(int, i.split('-'))) for i in value.split(',')]
        return FlashcardMask(mask)

    @classmethod
    def _psql_parse_mask(cls, value):
        return FlashcardMask(value)
