from datetime import datetime

from django.test import TestCase
from flashcards.models import User, Section, Flashcard, UserFlashcard, UserFlashcardQuiz
from flashcards.validators import FlashcardMask, OverlapIntervalException
from flashcards.serializers import QuizRequestSerializer, QuizResponseSerializer, QuizAnswerRequestSerializer
from flashy.settings import QUARTER_START, QUARTER_END


class RegistrationTests(TestCase):
    def setUp(self):
        User.objects.create_user(email="none@none.com", password="1234")

    def test_email_confirmation(self):
        user = User.objects.get(email="none@none.com")
        self.assertFalse(user.is_confirmed)
        user.confirm_email(user.confirmation_key)
        self.assertTrue(user.is_confirmed)


class UserTests(TestCase):
    def setUp(self):
        User.objects.create_user(email="none@none.com", password="1234")
        Section.objects.create(department='dept',
                               course_num='101a',
                               course_title='how 2 test',
                               instructor='George Lucas',
                               quarter='SP15')

    def test_section_list(self):
        section = Section.objects.get(course_num='101a')
        user = User.objects.get(email="none@none.com")
        self.assertNotIn(section, user.sections.all())
        user.sections.add(section)
        self.assertIn(section, user.sections.all())
        user.sections.add(section)
        self.assertEqual(user.sections.count(), 1)
        user.sections.remove(section)
        self.assertEqual(user.sections.count(), 0)


class FlashcardMaskTest(TestCase):
    def test_empty(self):
        try:
            fm = FlashcardMask([])
            self.assertEqual(fm.max_offset(), -1)
        except TypeError:
            self.fail()
        try:
            fm = FlashcardMask('')
            self.assertEqual(fm.max_offset(), -1)
        except TypeError:
            self.fail()
        try:
            fm = FlashcardMask(None)
            self.assertEqual(fm.max_offset(), -1)
        except TypeError:
            self.fail()

    def test_iterable(self):
        try:
            FlashcardMask(1)
        except TypeError as te:
            self.assertEqual(te.message, "Interval not a valid iterable")
        try:
            FlashcardMask([1, 2, 4])
        except TypeError as te:
            self.assertEqual(te.message, "Interval not a valid iterable")

    def test_interval(self):
        try:
            FlashcardMask([[1, 2, 3], [1]])
        except TypeError as te:
            self.assertEqual(te.message, "Intervals must have exactly 2 elements, begin and end")
        try:
            FlashcardMask([[1, 2], [1, 2, 4]])
        except TypeError as te:
            self.assertEqual(te.message, "Intervals must have exactly 2 elements, begin and end")
        try:
            FlashcardMask(([1, 2], [1]))
        except TypeError as te:
            self.assertEqual(te.message, "Intervals must have exactly 2 elements, begin and end")
        try:
            FlashcardMask("[1,2,3]")
        except TypeError as te:
            self.assertEqual(te.message, "Intervals must have exactly 2 elements, begin and end")

    def test_overlap(self):
        try:
            FlashcardMask({(1, 2), (2, 5)})
        except OverlapIntervalException as oie:
            self.assertEqual(oie.message, "Invalid interval offsets in the mask")
        try:
            FlashcardMask({(1, 20), (12, 15)})
        except OverlapIntervalException as oie:
            self.assertEqual(oie.message, "Invalid interval offsets in the mask")
        try:
            FlashcardMask({(2, 1), (5, 2)})
        except OverlapIntervalException as oie:
            self.assertEqual(oie.message, "Invalid interval offsets in the mask")


class FlashcardTests(TestCase):
    def setUp(self):
        section = Section.objects.create(department='dept',
                                         course_num='101a',
                                         course_title='how 2 test',
                                         instructor='George Lucas',
                                         quarter='SP15')
        user = User.objects.create_user(email="none@none.com", password="1234")
        user.sections.add(section)
        flashcard = Flashcard.objects.create(text="This is the text of the Flashcard",
                                             section=section,
                                             author=user,
                                             material_date=datetime.now(),
                                             previous=None,
                                             mask={(24, 34), (0, 4)})
        user.save()
        section.save()
        flashcard.save()

    def test_flashcard_edit(self):
        user = User.objects.get(email="none@none.com")
        user2 = User.objects.create_user(email="wow@wow.com", password="wow")
        section = Section.objects.get(course_title='how 2 test')
        user2.sections.add(section)
        user2.save()
        flashcard = Flashcard.objects.filter(author=user).get(text="This is the text of the Flashcard")
        pk_backup = flashcard.pk
        self.assertTrue(user.is_in_section(section))
        flashcard.edit(user, {})
        self.assertIsNotNone(flashcard.pk)
        UserFlashcard.objects.create(user=user2, flashcard=flashcard).save()
        flashcard = flashcard.edit(user2, {'text': 'This is the new text'})

        self.assertNotEqual(flashcard.pk, pk_backup)
        self.assertEqual(flashcard.text, 'This is the new text')

    def test_mask_field(self):
        user = User.objects.get(email="none@none.com")
        flashcard = Flashcard.objects.filter(author=user).get(text="This is the text of the Flashcard")
        self.assertTrue(isinstance(flashcard.mask, set))
        self.assertTrue(all([isinstance(interval, tuple) for interval in flashcard.mask]))
        blank1, blank2 = sorted(list(flashcard.mask))
        self.assertEqual(flashcard.text[slice(*blank1)], 'This')
        self.assertEqual(flashcard.text[slice(*blank2)], 'Flashcard')
        try:
            flashcard.mask = {(10, 34), (0, 14)}
            flashcard.save()
            self.fail()
        except OverlapIntervalException:
            self.assertTrue(True)


class UserFlashcardQuizTests(TestCase):
    def setUp(self):
        self.section = Section.objects.create(department='dept',
                                              course_num='101a',
                                              course_title='how 2 test',
                                              instructor='George Lucas',
                                              quarter='SP15')
        self.user = User.objects.create_user(email="none@none.com", password="1234")
        self.user.sections.add(self.section)
        self.flashcard = Flashcard.objects.create(text="This is the text of the Flashcard",
                                                  section=self.section,
                                                  author=self.user,
                                                  material_date=datetime.now(),
                                                  previous=None,
                                                  mask=[(24, 33), (0, 4)])
        self.user.save()
        self.section.save()
        self.flashcard.save()
        self.user_flashcard = UserFlashcard.objects.create(flashcard=self.flashcard,
                                                           user=self.user,
                                                           mask=self.flashcard.mask,
                                                           pulled=datetime.now())
        self.user_flashcard.save()
        self.user_flashcard.refresh_from_db()
        self.flashcard.refresh_from_db()

    def test_quiz_request(self):
        data = {'sections': [1], 'material_date_begin': QUARTER_START, 'material_date_end': QUARTER_END}
        serializer = QuizRequestSerializer(data=data)
        serializer.is_valid(raise_exception=True)
        validated_data = serializer.create(serializer.validated_data)
        user_flashcard = UserFlashcard.objects.filter(user=self.user,
                                                      flashcard__section__pk__in=validated_data['sections'],
                                                      flashcard__material_date__gte=validated_data['material_date_begin'],
                                                      flashcard__material_date__lte=validated_data['material_date_end'])
        self.assertTrue(user_flashcard.exists())

        user_flashcard = user_flashcard.first()
        self.assertEqual(user_flashcard, self.user_flashcard)

        mask = user_flashcard.mask.get_random_blank()
        word = user_flashcard.flashcard.text[slice(*mask)]
        user_flashcard_quiz = UserFlashcardQuiz(user_flashcard=user_flashcard,
                                                blanked_word=word)
        user_flashcard_quiz.save()
        self.assertTrue(isinstance(user_flashcard_quiz, UserFlashcardQuiz))
        self.assertIn(mask, [[24, 33], [0, 4]])

        user_flashcard_quiz.blanked_word = user_flashcard_quiz.user_flashcard.flashcard.text[slice(*mask)]
        self.assertIn(user_flashcard_quiz.blanked_word, ["This", "Flashcard"])
        user_flashcard_quiz.save()

        response = QuizResponseSerializer(instance=user_flashcard_quiz, mask=mask).data
        self.assertEqual(response['pk'], 1)
        self.assertEqual(response['section'], 1)
        self.assertEqual(response['text'], user_flashcard_quiz.user_flashcard.flashcard.text)
        self.assertEqual(response['mask'], mask)

    def test_quiz_answer(self):
        data = {'response': 'Flashcard'}
        mask = self.user_flashcard.mask.get_random_blank()
        word = self.flashcard.text[slice(*mask)]
        user_flashcard_quiz = UserFlashcardQuiz(user_flashcard=self.user_flashcard, blanked_word=word)
        user_flashcard_quiz.save()
        serializer = QuizAnswerRequestSerializer(instance=user_flashcard_quiz, data=data)
        serializer.is_valid()
        serializer.update(user_flashcard_quiz, serializer.validated_data)
        self.assertEqual(user_flashcard_quiz.response, data['response'])

        data = {'correct': True}
        serializer = QuizAnswerRequestSerializer(instance=user_flashcard_quiz, data=data)
        serializer.is_valid()
        serializer.update(user_flashcard_quiz, serializer.validated_data)
        self.assertTrue(user_flashcard_quiz.correct)
