from datetime import datetime

import os
from pytz import UTC

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

IN_PRODUCTION = 'FLASHY_PRODUCTION' in os.environ

DEBUG = not IN_PRODUCTION

ALLOWED_HOSTS = ['127.0.0.1', 'flashy.cards']

if DEBUG:
    ABSOLUTE_URL_ROOT = 'http://127.0.0.1:8080/'
else:
    ABSOLUTE_URL_ROOT = 'https://flashy.cards/'

AUTH_USER_MODEL = 'flashcards.User'
REST_FRAMEWORK = {
    'DEFAULT_PAGINATION_CLASS': 'rest_framework.pagination.LimitOffsetPagination',
    'PAGE_SIZE': 20,
    'DEFAULT_THROTTLE_CLASSES': (
        'rest_framework.throttling.UserRateThrottle',
    ),
}

INSTALLED_APPS = [
    'simple_email_confirmation',
    'flashcards',
    'django.contrib.admin',
    'django.contrib.admindocs',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'ws4redis',
    'rest_framework_swagger',
    'rest_framework',
    'django_extensions',
]

WEBSOCKET_URL = '/ws/'

MIDDLEWARE_CLASSES = (
    'flashcards.middleware.SetRemoteAddrFromForwardedFor',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.auth.middleware.SessionAuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'django.middleware.security.SecurityMiddleware',

)

ROOT_URLCONF = 'flashy.urls'

AUTHENTICATION_BACKENDS = (
    'django.contrib.auth.backends.ModelBackend',
)

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': ['templates/'],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
                'django.core.context_processors.static',
                'ws4redis.context_processors.default',
            ],
        },
    },
]

WSGI_APPLICATION = 'ws4redis.django_runserver.application'

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}

if IN_PRODUCTION:
    DATABASES['default'] = {
        'ENGINE': 'django.db.backends.postgresql_psycopg2',
        'NAME': 'flashy',
        'USER': 'flashy',
        'PASSWORD': os.environ['FLASHY_DB_PW'],
        'HOST': 'localhost',
        'PORT': '',
    }

LANGUAGE_CODE = 'en-us'
USE_I18N = True
USE_L10N = True
USE_TZ = True
TIME_ZONE = 'America/Los_Angeles'

QUARTER_START = UTC.localize(datetime(2015, 3, 30))
QUARTER_END = UTC.localize(datetime(2015, 7, 12))

STATIC_URL = '/static/'
STATIC_ROOT = 'static'

# Four settings just to be sure
EMAIL_FROM = 'noreply@flashy.cards'
EMAIL_HOST_USER = 'noreply@flashy.cards'
DEFAULT_FROM_EMAIL = 'noreply@flashy.cards'
SERVER_EMAIL = 'noreply@flashy.cards'

EMAIL_BACKEND = 'django.core.mail.backends.console.EmailBackend'

if DEBUG:
    INSTALLED_APPS.append('debug_toolbar')
    # DEBUG_TOOLBAR_PANELS = ['debug_toolbar.panels.profiling.ProfilingPanel']

if IN_PRODUCTION:
    INSTALLED_APPS.append('django_ses')
    AWS_SES_REGION_NAME = 'us-west-2'
    AWS_SES_REGION_ENDPOINT = 'email.us-west-2.amazonaws.com'
    EMAIL_BACKEND = 'django_ses.SESBackend'

if IN_PRODUCTION:
    SESSION_COOKIE_SECURE = True
    CSRF_COOKIE_SECURE = True
    SECURE_PROXY_SSL_HEADER = ('HTTP_X_FORWARDED_PROTO', 'https')
    # are we secure yet?

if IN_PRODUCTION:
    LOGGING = {
        'version': 1,
        'disable_existing_loggers': False,
        'handlers': {
            'file': {
                'level': 'DEBUG',
                'class': 'logging.FileHandler',
                'filename': 'debug.log',
            },
            'eventslog': {
                'level': 'INFO',
                'class': 'logging.FileHandler',
                'filename': 'events.log',
                'formatter': 'verbose'
            },
        },
        'formatters': {
            'verbose': {
                'format': '%(asctime)s %(module)s %(message)s'
            },
        },
        'loggers': {
            'django': {
                'handlers': ['file'],
                'level': 'DEBUG',
                'propagate': True,
            },
            'flashy.events': {
                'level': 'INFO',
                'propagate': True,
                'handlers': ['eventslog'],
            }
        },
    }


# If we're in production, SECRET_KEY is pulled from the environ
# Otherwise it doesn't matter
SECRET_KEY = os.environ.get('SECRET_KEY', 'LOL DEFAULT SECRET KEY')
GCM_API_KEY = os.environ.get('GCM_API_KEY', 'GCM PLS')

CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.locmem.LocMemCache'
    }
}

SWAGGER_SETTINGS = {
    'doc_expansion': 'list'
}

WS4REDIS_EXPIRE = 0
SESSION_ENGINE = 'redis_sessions.session'

FEED_PAGE_SIZE = 50